@@ -23,23 +23,106 @@ from contextlib import contextmanager
 import torch
 
 
-def stabilize(input, epsilon=1e-6):
-    '''Stabilize input for safe division. This shifts zero-elements by ``+ epsilon``. For the sake of the
-    *epsilon rule*, this also shifts positive values by ``+ epsilon`` and negative values by ``- epsilon``.
+class Stabilizer:
+    '''Class to create a stabilizer callable.
+
+    Parameters
+    ----------
+    epsilon: float, optional
+        Value by which to shift/clip elements of ``input``.
+    clip: bool, optional
+        If ``False`` (default), add ``epsilon`` multiplied by each entry's sign (+1 for 0). If ``True``, instead clip
+        the absolute value of ``input`` and multiply it by each entry's original sign.
+    norm_scale: bool, optional
+        If ``False`` (default), ``epsilon`` is added to/used to clip ``input``. If ``True``, scale ``epsilon`` by the
+        square root of the mean over the squared elements of the specified dimensions ``dim``.
+    dim: tuple[int], optional
+        If ``norm_scale`` is ``True``, specifies the dimension over which the scaled norm should be computed (all
+        except dimension 0 by default).
+
+    '''
+    def __init__(self, epsilon=1e-6, clip=False, norm_scale=False, dim=None):
+        self.epsilon = epsilon
+        self.clip = clip
+        self.norm_scale = norm_scale
+        self.dim = dim
+
+    def __call__(self, input):
+        '''Stabilize input for safe division. This shifts zero-elements by ``+ epsilon``. For the sake of the
+        *epsilon rule*, this also shifts positive values by ``+ epsilon`` and negative values by ``- epsilon``.
+
+        Parameters
+        ----------
+        input: :py:obj:`torch.Tensor`
+            Tensor to stabilize.
+
+        Returns
+        -------
+        :py:obj:`torch.Tensor`
+            Stabilized ``input``.
+        '''
+        return stabilize(input, self.epsilon, self.clip, self.norm_scale, self.dim)
+
+    @classmethod
+    def ensure(cls, value):
+        '''Given a value, return a stabilizer. If ``value`` is a float, a Stabilizer with that epsilon ``value`` is
+        returned. If ``value`` is callable, it will be used directly as a stabilizer. Otherwise a TypeError will be
+        raised.
+
+        Parameters
+        ----------
+        value: float, int, or callable
+            The value used to produce a valid stabilizer function.
+
+        Returns
+        -------
+        callable or Stabilizer
+            A callable to be used as a stabilizer.
+
+        Raises
+        ------
+        TypeError
+            If no valid stabilizer could be produced from ``value``.
+        '''
+        if isinstance(value, (float, int)):
+            return cls(epsilon=float(value))
+        if callable(value):
+            return value
+        raise TypeError(f'Value {value} is not a valid stabilizer!')
+
+
+def stabilize(input, epsilon=1e-6, clip=False, norm_scale=False, dim=None):
+    '''Stabilize input for safe division.
 
     Parameters
     ----------
     input: :py:obj:`torch.Tensor`
         Tensor to stabilize.
     epsilon: float, optional
-        Value by which to shift elements.
+        Value by which to shift/clip elements of ``input``.
+    clip: bool, optional
+        If ``False`` (default), add ``epsilon`` multiplied by each entry's sign (+1 for 0). If ``True``, instead clip
+        the absolute value of ``input`` and multiply it by each entry's original sign.
+    norm_scale: bool, optional
+        If ``False`` (default), ``epsilon`` is added to/used to clip ``input``. If ``True``, scale ``epsilon`` by the
+        square root of the mean over the squared elements of the specified dimensions ``dim``.
+    dim: tuple[int], optional
+        If ``norm_scale`` is ``True``, specifies the dimension over which the scaled norm should be computed. Defaults
+        to all except dimension 0.
 
     Returns
     -------
     :py:obj:`torch.Tensor`
         New Tensor copied from `input` with values shifted by epsilon.
     '''
-    return input + ((input == 0.).to(input) + input.sign()) * epsilon
+    sign = ((input == 0.).to(input) + input.sign())
+    if norm_scale:
+        if dim is None:
+            dim = tuple(range(1, input.ndim))
+        epsilon = epsilon * ((input ** 2).mean(dim=dim, keepdim=True) ** .5)
+    if clip:
+        return sign * input.abs().clip(min=epsilon)
+    return input + sign * epsilon
 
 
 def expand(tensor, shape, cut_batch_dim=False):
@@ -92,54 +175,161 @@ def expand(tensor, shape, cut_batch_dim=False):
     raise RuntimeError('Invalid shape! Target: {tensor.shape}; Source: {shape}')
 
 
-@contextmanager
-def mod_params(module, modifier, param_keys=None, require_params=True):
-    '''Context manager to temporarily modify parameter attributes (all by default) of a module.
+def zero_wrap(zero_params):
+    '''Create a function wrapper factory (i.e. a decorator), which takes a single function argument ``(name, param) ->
+    tensor`` such that the function is only called if name is not equal to zero_params, if zero_params is a string, or
+    it is not in zero_params. Otherwise return `torch.zeros_like` of that tensor.
+
+    Parameters
+    ----------
+    zero_params: str or list[str]
+        String or list of strings compared to `name`.
+
+    Returns
+    -------
+    function
+        The function wrapper to be called on the function.
+    '''
+    def zero_params_wrapper(modifier):
+        '''Wrap a function (name, param) -> tensor such that the function is only called if name is not equal to the
+        closure list zero_params, if zero_params is a string, or it is not in zero_params. Otherwise return
+        `torch.zeros_like` of that tensor.
+
+        Parameters
+        ----------
+        func: function
+            Function to wrap.
+
+        Returns
+        -------
+        function
+            The wrapped function.
+        '''
+        if not zero_params:
+            return modifier
+
+        @functools.wraps(modifier)
+        def modifier_wrapper(input, name):
+            '''Wrapped function (name, param) -> tensor, where the original function is only called if name is not
+            equal to the closure list zero_params, if zero_params is a string, or it is not in zero_params. Otherwise
+            return `torch.zeros_like` of that tensor.
+
+            Parameters
+            ----------
+            input: :py:obj:`torch.Tensor`
+                The input tensor modified by the original function.
+            name: str
+                The name associated with the input tensor (e.g. the parameter name).
+
+            Returns
+            -------
+            :py:obj:`torch.Tensor`
+                The modified tensor.
+            '''
+            if isinstance(zero_params, str) and name == zero_params or name in zero_params:
+                return torch.zeros_like(input)
+            return modifier(input, name)
+
+        return modifier_wrapper
+    return zero_params_wrapper
+
+
+class ParamMod:
+    '''Class to produce a context manager to temporarily modify parameter attributes (all by default) of a module.
 
     Parameters
     ----------
-    module: obj:`torch.nn.Module`
-        Module of which to modify parameters. If `requires_params` is `True`, it must have all elements given in
-        `param_keys` as attributes (attributes are allowed to be `None`, in which case they are ignored).
     modifier: function
         A function used to modify parameter attributes. If `param_keys` is empty, this is not used.
     param_keys: list[str], optional
-        A list of parameters that shall be modified. If `None` (default), all parameters are modified (which may be
-        none). If `[]`, no parameters are modified and `modifier` is ignored.
+        A list of parameter names that shall be modified. If `None` (default), all parameters are modified (which may
+        be none). If `[]`, no parameters are modified and `modifier` is ignored.
+    zero_params: list[str], optional
+        A list of parameter names that shall set to zero. If `None` (default), no parameters are set to zero.
     require_params: bool, optional
         Whether existence of `module`'s params is mandatory (True by default). If the attribute exists but is `None`,
         it is not considered missing, and the modifier is not applied.
-
-    Raises
-    ------
-    RuntimeError
-        If `require_params` is `True` and `module` is missing an attribute listed in `param_keys`.
-
-    Yields
-    ------
-    module: obj:`torch.nn.Module`
-        The `module` with appropriate parameters temporarily modified.
     '''
-    try:
-        stored_params = {}
-        if param_keys is None:
-            param_keys = [name for name, _ in module.named_parameters(recurse=False)]
+    def __init__(self, modifier, param_keys=None, zero_params=None, require_params=True):
+        self.modifier = modifier
+        self.param_keys = param_keys
+        self.zero_params = zero_params
+        self.require_params = require_params
 
-        missing = [key for key in param_keys if not hasattr(module, key)]
-        if require_params and missing:
-            missing_str = '\', \''.join(missing)
-            raise RuntimeError(f'Module {module} requires missing parameters: \'{missing_str}\'')
+    @classmethod
+    def ensure(cls, modifier):
+        '''If ``modifier`` is an instance of ParamMod, return it as-is, if it is callable, create a new instance with
+        ``modifier`` as the ParamMod's function, otherwise raise a TypeError.
 
-        for key in param_keys:
-            if key not in missing:
-                param = getattr(module, key)
-                if param is not None:
-                    stored_params[key] = param
-                    setattr(module, key, torch.nn.Parameter(modifier(param.data, key)))
-        yield module
-    finally:
-        for key, value in stored_params.items():
-            setattr(module, key, value)
+        Parameters
+        ----------
+        modifier : :py:obj:`ParamMod` or callable
+            The modifier which, if necessary, will be used to construct a ParamMod.
+
+        Returns
+        -------
+        :py:obj:`ParamMod`
+            Either ``modifier`` as is, or a :py:obj:`ParamMod` constructed using ``modifier``.
+
+        Raises
+        ------
+        TypeError
+            If ``modifier`` is neither an instance of :py:obj:`ParamMod`, nor callable.
+        '''
+        if isinstance(modifier, cls):
+            return modifier
+        if callable(modifier):
+            return cls(modifier)
+        raise TypeError(f'{modifier} is neither an instance of {cls}, nor callable!')
+
+    @contextmanager
+    def __call__(self, module):
+        '''Context manager to temporarily modify parameter attributes (all by default) of a module.
+
+        Parameters
+        ----------
+        module: :py:obj:`torch.nn.Module`
+            Module of which to modify parameters. If `self.requires_params` is `True`, it must have all elements given
+            in `self.param_keys` as attributes (attributes are allowed to be `None`, in which case they are ignored).
+
+        Raises
+        ------
+        RuntimeError
+            If `self.require_params` is `True` and `module` is missing an attribute listed in `self.param_keys`.
+
+        Yields
+        ------
+        module: :py:obj:`torch.nn.Module`
+            The `module` with appropriate parameters temporarily modified.
+        '''
+        try:
+            stored_params = {}
+            param_keys = self.param_keys
+            zero_params = self.zero_params
+
+            if param_keys is None:
+                param_keys = [name for name, _ in module.named_parameters(recurse=False)]
+            if zero_params is None:
+                zero_params = []
+
+            missing = [key for key in param_keys if not hasattr(module, key)]
+            if self.require_params and missing:
+                missing_str = '\', \''.join(missing)
+                raise RuntimeError(f'Module {module} requires missing parameters: \'{missing_str}\'')
+
+            modifier = zero_wrap(zero_params)(self.modifier)
+
+            for key in param_keys:
+                if key not in missing:
+                    param = getattr(module, key)
+                    if param is not None:
+                        stored_params[key] = param
+                        setattr(module, key, torch.nn.Parameter(modifier(param.data, key)))
+
+            yield module
+        finally:
+            for key, value in stored_params.items():
+                setattr(module, key, value)
 
 
 def collect_leaves(module):
@@ -147,12 +337,12 @@ def collect_leaves(module):
 
     Parameters
     ----------
-    module: obj:`torch.nn.Module`
+    module: :py:obj:`torch.nn.Module`
         A module for which the leaves will be collected.
 
     Yields
     ------
-    leaf: obj:`torch.nn.Module`
+    leaf: :py:obj:`torch.nn.Module`
         Either a leaf of the module structure, or the module itself if it has no children.
     '''
     is_leaf = True
@@ -253,35 +443,34 @@ class Hook:
 
 
 class BasicHook(Hook):
-    '''A hook to compute the layerwise attribution of the module it is attached to.
-    A `BasicHook` instance may only be registered with a single module.
+    '''A hook to compute the layer-wise attribution of the module it is attached to.
+    A BasicHook instance may only be registered with a single module.
 
     Parameters
     ----------
     input_modifiers: list[callable], optional
-        A list of functions to produce multiple inputs. Default is a single input which is the identity.
-    param_modifiers: list[callable], optional
-        A list of functions to temporarily modify the parameters of the attached module for each input produced
-        with `input_modifiers`. Default is unmodified parameters for each input.
+        A list of functions ``(input: torch.Tensor) -> torch.Tensor`` to produce multiple inputs. Default is a single
+        input which is the identity.
+    param_modifiers: list[:py:obj:`~zennit.core.ParamMod` or callable], optional
+        A list of ParamMod instances or functions ``(obj: torch.Tensor, name: str) -> torch.Tensor``, with parameter
+        tensor ``obj``, registered in the root model as ``name``, to temporarily modify the parameters of the attached
+        module for each input produced with `input_modifiers`. Default is unmodified parameters for each input. Use a
+        :py:obj:`~zennit.core.ParamMod` instance to specify which parameters should be modified, whether they are
+        required, and which should be set to zero.
     output_modifiers: list[callable], optional
-        A list of functions to modify the module's output computed using the modified parameters before gradient
-        computation for each input produced with `input_modifier`. Default is the identity for each output.
+        A list of functions ``(input: torch.Tensor) -> torch.Tensor`` to modify the module's output computed using the
+        modified parameters before gradient computation for each input produced with `input_modifier`. Default is the
+        identity for each output.
     gradient_mapper: callable, optional
-        Function to modify upper relevance. Call signature is of form `(grad_output, outputs)` and a tuple of
-        the same size as outputs is expected to be returned. `outputs` has the same size as `input_modifiers` and
-        `param_modifiers`. Default is a stabilized normalization by each of the outputs, multiplied with the output
-        gradient.
+        Function ``(out_grad: torch.Tensor, outputs: list[torch.Tensor]) -> list[torch.Tensor]`` to modify upper
+        relevance. A list or tuple of the same size as ``outputs`` is expected to be returned. ``outputs`` has the same
+        size as ``input_modifiers`` and ``param_modifiers``. Default is a stabilized normalization by each of the
+        outputs, multiplied with the output gradient.
     reducer: callable
-        Function to reduce all the inputs and gradients produced through `input_modifiers` and `param_modifiers`.
-        Call signature is of form `(inputs, gradients)`, where `inputs` and `gradients` have the same as
-        `input_modifiers` and `param_modifiers`. Default is the sum of the multiplications of each input and its
-        corresponding gradient.
-    param_keys: list[str], optional
-        A list of parameters that shall be modified. If `None` (default), all parameters are modified (which may be
-        none). If `[]`, no parameters are modified and `modifier` is ignored.
-    require_params: bool, optional
-        Whether existence of `module`'s params is mandatory (True by default). If the attribute exists but is `None`,
-        it is not considered missing, and the modifier is not applied.
+        Function ``(inputs: list[torch.Tensor], gradients: list[torch.Tensor]) -> torch.Tensor`` to reduce all the
+        inputs and gradients produced through ``input_modifiers`` and ``param_modifiers``. ``inputs`` and ``gradients``
+        have the same as ``input_modifiers`` and ``param_modifiers``. Default is the sum of the multiplications of each
+        input and its corresponding gradient.
     '''
     def __init__(
         self,
@@ -290,8 +479,7 @@ class BasicHook(Hook):
         output_modifiers=None,
         gradient_mapper=None,
         reducer=None,
-        param_keys=None,
-        require_params=True
+        stabilizer=1e-6,
     ):
         super().__init__()
         modifiers = {
@@ -314,9 +502,6 @@ class BasicHook(Hook):
         self.gradient_mapper = gradient_mapper
         self.reducer = reducer
 
-        self.param_keys = param_keys
-        self.require_params = require_params
-
     def forward(self, module, input, output):
         '''Forward hook to save module in-/outputs.'''
         self.stored_tensors['input'] = input
@@ -324,18 +509,17 @@ class BasicHook(Hook):
     def backward(self, module, grad_input, grad_output):
         '''Backward hook to compute LRP based on the class attributes.'''
         original_input = self.stored_tensors['input'][0].detach()
-        param_kwargs = dict(param_keys=self.param_keys, require_params=self.require_params)
         inputs = []
         outputs = []
         for in_mod, param_mod, out_mod in zip(self.input_modifiers, self.param_modifiers, self.output_modifiers):
             input = in_mod(original_input).requires_grad_()
-            with mod_params(module, param_mod, **param_kwargs) as modified, torch.autograd.enable_grad():
+            with ParamMod.ensure(param_mod)(module) as modified, torch.autograd.enable_grad():
                 output = modified.forward(input)
                 output = out_mod(output)
             inputs.append(input)
             outputs.append(output)
-        gradients = torch.autograd.grad(outputs, inputs, grad_outputs=self.gradient_mapper(grad_output[0], outputs))
-        # relevance = self.reducer([input.detach() for input in inputs], [gradient.detach() for gradient in gradients])
+        grad_outputs = self.gradient_mapper(grad_output[0], outputs)
+        gradients = torch.autograd.grad(outputs, inputs, grad_outputs=grad_outputs)
         relevance = self.reducer(inputs, gradients)
         return tuple(relevance if original.shape == relevance.shape else None for original in grad_input)
 
@@ -343,15 +527,16 @@ class BasicHook(Hook):
         '''Return a copy of this hook.
         This is used to describe hooks of different modules by a single hook instance.
         '''
-        return BasicHook(
+        copy = BasicHook.__new__(type(self))
+        BasicHook.__init__(
+            copy,
             self.input_modifiers,
             self.param_modifiers,
             self.output_modifiers,
             self.gradient_mapper,
             self.reducer,
-            self.param_keys,
-            self.require_params
         )
+        return copy
 
     @staticmethod
     def _default_modifier(obj, name=None):
