@@ -3,26 +3,100 @@ from itertools import product
 
 import torch
 import pytest
-from helpers import nograd
+from helpers import nograd, prodict
 
-from zennit.core import stabilize, expand, mod_params, collect_leaves
+from zennit.core import stabilize, expand, ParamMod, collect_leaves, Stabilizer
 from zennit.core import Identity, Hook, BasicHook, RemovableHandle, RemovableHandleList, Composite
 
 
-@pytest.mark.parametrize('input,epsilon,expected', [
-    ([0., -0., 1., -1.], None, [1e-6, 1e-6, 1. + 1e-6, -1. - 1e-6]),
-    ([0., -0., 1., -1.], 1e-3, [1e-3, 1e-3, 1. + 1e-3, -1. - 1e-3]),
-    ([0., -0., 1., -1.], 1., [1., 1., 2., -2.]),
+@pytest.mark.parametrize('kwargs,input,expected', [
+    (
+        {},
+        [0., -0., 1., -1.],
+        [1e-6, 1e-6, 1. + 1e-6, -1. - 1e-6]
+    ), (
+        {'epsilon': 1e-3, 'clip': False, 'norm_scale': False},
+        [0., -0., 1., -1.],
+        [1e-3, 1e-3, 1. + 1e-3, -1. - 1e-3]
+    ), (
+        {'epsilon': 1., 'clip': False, 'norm_scale': False},
+        [0., -0., 1., -1.],
+        [1., 1., 2., -2.]
+    ), (
+        {'epsilon': 1e-6, 'clip': True, 'norm_scale': False},
+        [0., -0., 1., -1.],
+        [1e-6, 1e-6, 1., -1.]
+    ), (
+        {'epsilon': 1e-3, 'clip': True, 'norm_scale': False},
+        [0., -0., 1., -1.],
+        [1e-3, 1e-3, 1., -1.]
+    ), (
+        {'epsilon': 1., 'clip': True, 'norm_scale': False},
+        [0., -0., 1., -1.],
+        [1., 1., 1., -1.]
+    ), (
+        {'epsilon': 1e-6, 'clip': False, 'norm_scale': True},
+        [0., -0., 2., -2.],
+        [1.4142e-6, 1.4142e-6, 2. + 1.4142e-6, -2. - 1.4142e-6]
+    ), (
+        {'epsilon': 1e-3, 'clip': False, 'norm_scale': True},
+        [0., -0., 2., -2.],
+        [1.4142e-3, 1.4142e-3, 2. + 1.4142e-3, -2. - 1.4142e-3]
+    ), (
+        {'epsilon': 1., 'clip': False, 'norm_scale': True},
+        [0., -0., 2., -2.],
+        [1.4142, 1.4142, 3.4142, -3.4142]
+    ), (
+        {'epsilon': 1e-6, 'clip': True, 'norm_scale': True},
+        [0., -0., 2., -2.],
+        [1.4142e-6, 1.4142e-6, 2., -2.]
+    ), (
+        {'epsilon': 1e-3, 'clip': True, 'norm_scale': True},
+        [0., -0., 2., -2.],
+        [1.4142e-3, 1.4142e-3, 2., -2.]
+    ), (
+        {'epsilon': 1., 'clip': True, 'norm_scale': True},
+        [0., -0., 2., -2.],
+        [1.4142, 1.4142, 2., -2.]
+    ),
 ])
-def test_stabilize(input, epsilon, expected):
+def test_stabilize(kwargs, input, expected):
     '''Test whether stabilize produces the expected outputs given some inputs.'''
-    kwargs = {} if epsilon is None else {'epsilon': epsilon}
     input_tensor = torch.tensor(input, dtype=torch.float64)
-    output = stabilize(input_tensor, **kwargs)
+    output = stabilize(input_tensor, dim=0, **kwargs)
     expected_tensor = torch.tensor(expected, dtype=torch.float64)
     assert torch.allclose(expected_tensor, output)
 
 
+@pytest.mark.parametrize('kwargs', prodict(
+    epsilon=[1e-6, 1e-3, 1.],
+    clip=[True, False],
+    norm_scale=[True, False],
+    dim=[None, (0,), (1,), (0, 1)]
+))
+def test_stabilizer_match(kwargs, data_simple):
+    '''Test whether stabilize and Stabilizer produce the same output.'''
+    stabilizer = Stabilizer(**kwargs)
+    stabilizer_out = stabilizer(data_simple)
+    stabilize_out = stabilize(data_simple, **kwargs)
+    assert torch.allclose(stabilizer_out, stabilize_out)
+
+
+@pytest.mark.parametrize('value', [1., 1, Stabilizer(epsilon=1.), lambda x: x])
+def test_stabilizer_ensure(value):
+    '''Test whether Stabilizer.ensure produces a stabilizer with the correct epsilon, or returns callables as-is.'''
+    ensured = Stabilizer.ensure(value)
+    assert not isinstance(value, float) or isinstance(ensured, Stabilizer) and ensured.epsilon == value
+    assert not callable(value) or value is ensured
+
+
+@pytest.mark.parametrize('value', [None, 'wow'])
+def test_stabilizer_ensure_fail(value):
+    '''Test whether Stabilizer.ensure fails on unsupported types.'''
+    with pytest.raises(TypeError):
+        Stabilizer.ensure(value)
+
+
 @pytest.mark.parametrize('input_shape,target_shape,cut_batch_dim', [
     ((), (), False),
     ((), (2,), False),
@@ -60,12 +134,17 @@ def test_expand_shape_invalid(input_shape, target_shape):
         expand(input, target_shape)
 
 
-@pytest.mark.parametrize('param_keys,require_params,bias', [
-    *product([None, ['weight'], ['bias'], ['weight', 'bias']], [True], [False, True]),
-    *product([['weight', 'bias', 'beight', 'wias']], [False], [False, True]),
+@pytest.mark.parametrize('param_keys,require_params,zero_params,bias', [
+    *product(
+        [None, ['weight'], ['bias'], ['weight', 'bias']],
+        [True],
+        [[], ['weight'], ['bias'], ['weight', 'bias']],
+        [False, True]
+    ),
+    *product([['weight', 'bias', 'beight', 'wias']], [False], [[]], [False, True]),
 ])
-def test_mod_params(param_keys, require_params, bias):
-    '''Test whether mod_params correctly changes and restores parameters of a module.'''
+def test_param_mod(param_keys, require_params, zero_params, bias):
+    '''Test whether ParamMod correctly changes and restores parameters of a module.'''
     module = nograd(torch.nn.Linear(2, 2, bias=bias))
     mod_param_keys = param_keys
     if param_keys is None:
@@ -73,29 +152,54 @@ def test_mod_params(param_keys, require_params, bias):
 
     for key in param_keys:
         if getattr(module, key, None) is not None:
-            getattr(module, key).data = torch.zeros_like(getattr(module, key))
+            getattr(module, key).data = torch.full_like(getattr(module, key), 0.5)
 
-    with mod_params(
-        module, lambda x, _: torch.ones_like(x), param_keys=mod_param_keys, require_params=require_params
-    ) as modified:
+    param_mod = ParamMod(
+        lambda x, _: torch.ones_like(x),
+        param_keys=mod_param_keys,
+        require_params=require_params,
+        zero_params=zero_params,
+    )
+
+    with param_mod(module) as modified:
         for key in param_keys:
             if getattr(module, key, None) is not None:
-                assert torch.all(getattr(modified, key) == 1.0), f'Parameter \'{key}\' was not modified!'
+                if key not in zero_params:
+                    assert torch.all(getattr(modified, key) == 1.0), f'Parameter \'{key}\' was not modified!'
+                else:
+                    assert torch.all(getattr(modified, key) == 0.0), f'Parameter \'{key}\' was not set to zero!'
 
     for key in param_keys:
         if getattr(module, key, None) is not None:
-            assert torch.all(getattr(module, key) == 0.0), f'Parameter \'{key}\' was not restored!'
+            assert torch.all(getattr(module, key) == 0.5), f'Parameter \'{key}\' was not restored!'
 
 
-def test_mod_params_required_missing():
-    '''Test wether mod_params raises a RuntimeError when it is missing parameters when require_params=True.'''
+def test_param_mod_required_missing():
+    '''Test wether ParamMod raises a RuntimeError when it is missing parameters when require_params=True.'''
     module = torch.nn.Module()
     param_keys = ['beight', 'wias']
+    param_mod = ParamMod(lambda x, _: torch.ones_like(x), param_keys=param_keys, require_params=True)
     with pytest.raises(RuntimeError):
-        with mod_params(module, lambda x, _: torch.ones_like(x), param_keys=param_keys, require_params=True):
+        with param_mod(module):
             pass
 
 
+@pytest.mark.parametrize('modifier', [lambda x: x, ParamMod(lambda x: x)])
+def test_param_mod_ensure(modifier):
+    '''Test wether ParamMod.ensure returns the original object if it is an instance of ParamMod, or a new instance of
+    ParamMod if it is not an instance of ParamMod and callable.'''
+    result = ParamMod.ensure(modifier)
+    assert isinstance(result, ParamMod)
+    assert not isinstance(modifier, ParamMod) or result is modifier
+
+
+def test_param_mod_ensure_unknown_type():
+    '''Test wether ParamMod.ensure raises a TypeError when it is called on a non-callable object that is not of type
+    ParamMod.'''
+    with pytest.raises(TypeError):
+        ParamMod.ensure(None)
+
+
 def test_collect_leaves_dummy():
     '''Test whether collect leaves correctly collects all leaves of an example model.'''
     modules = []
@@ -292,12 +396,16 @@ def test_basic_hook_custom():
 
     hook = BasicHook(
         input_modifiers=[assert_inputs],
-        param_modifiers=[assert_params],
+        param_modifiers=[
+            ParamMod(
+                assert_params,
+                param_keys=['weight', 'bias'],
+                require_params=True,
+            )
+        ],
         output_modifiers=[assert_outputs],
         gradient_mapper=assert_gradient,
         reducer=assert_reducer,
-        param_keys=['weight', 'bias'],
-        require_params=True
     )
 
     call_events = ('input', 'params', 'output', 'gradient', 'reducer')
@@ -322,12 +430,17 @@ def test_basic_hook_copy():
     '''Test whether BasicHook.copy copies the Hook correctly.'''
     hook = BasicHook(
         input_modifiers=[lambda obj, name: obj],
-        param_modifiers=[lambda obj, name: obj],
+        param_modifiers=[
+            ParamMod(
+                (lambda obj, name: obj),
+                param_keys=['weight', 'bias'],
+                require_params=True,
+                zero_params=['bias'],
+            ),
+        ],
         output_modifiers=[lambda obj, name: obj],
         gradient_mapper=(lambda out_grad, outputs: [out_grad for output in outputs]),
         reducer=(lambda inputs, gradients: [input * gradient for input, gradient in zip(inputs, gradients)]),
-        param_keys=['weight', 'bias'],
-        require_params=True
     )
     copy = hook.copy()
 
@@ -337,8 +450,6 @@ def test_basic_hook_copy():
         'output_modifiers',
         'reducer',
         'gradient_mapper',
-        'param_keys',
-        'require_params'
     )
 
     assert copy is not hook
